@@ -4,7 +4,8 @@ from typing import Optional, Text, List, Dict, Any, Union
 from functools import wraps
 
 import tensorflow_model_analysis as tfma
-from tfx.orchestration import metadata, pipeline, data_types
+from tfx.orchestration.pipeline import Pipeline
+from tfx.orchestration import metadata, data_types
 
 from tfx.proto import example_gen_pb2, trainer_pb2, \
     infra_validator_pb2, pusher_pb2, bulk_inferrer_pb2
@@ -56,21 +57,24 @@ def build_step(component_name: Text):
 
 class PipelineDef:
     """The fluent definition class that empowers fluent-tfx.
-    Initialise the constructor with a name and use the 
+    Initialise the constructor with a name and use the
     factory methods to construct the pipeline.
     Obtain the tfx pipeline instance by using `.build()`
     """
 
-    def __init__(self, name: Text, bucket: Optional[Text] = 'bucket',
+    def __init__(self, name: Text, bucket: Optional[Text] = None,
                  metadata_connection_config: Optional[metadata_store_pb2.ConnectionConfig] = None):
         """
         Args:
             name (Text): The pipeline's name
             bucket (Text, optional): Intermediate artifacts and staging binaries are going
             to be saved under `{name}/{bucket}/..`. Defaults to './bucket'.
-            metadata_connection_config (Optional[metadata_store_pb2.ConnectionConfig], optional): 
+            metadata_connection_config (Optional[metadata_store_pb2.ConnectionConfig], optional):
             Optional ML Metadata configuration for rapid local prototyping you can use `with_sqlite_ml_metadata`. Defaults to None.
         """
+        if not bucket:
+            bucket = os.path.join(os.getcwd(), 'bucket')
+
         self.components = {}
 
         self.pipeline_name = name
@@ -97,7 +101,7 @@ class PipelineDef:
         self.bulk_inferrer = None
 
     def with_sqlite_ml_metadata(self):
-        """WIll use a sqlite database under {bucket}/{name}/metadata.db as a 
+        """WIll use a sqlite database under {bucket}/{name}/metadata.db as a
         backend for ML Metadata.
 
         Returns: self
@@ -153,7 +157,7 @@ class PipelineDef:
         """
 
         self.example_gen = input_builders.from_csv(
-            uri, input_config=input_config, output_config=output_config)
+            uri=uri, input_config=input_config, output_config=output_config)
         return self.example_gen
 
     @build_step('example_gen')
@@ -170,7 +174,7 @@ class PipelineDef:
         Returns: self
         """
         self.example_gen = input_builders.from_tfrecord(
-            uri, input_config=input_config, output_config=output_config)
+            uri=uri, input_config=input_config, output_config=output_config)
         return self.example_gen
 
     @build_step('example_gen')
@@ -187,7 +191,7 @@ class PipelineDef:
         Returns: self
         """
         self.example_gen = input_builders.from_bigquery(
-            query, input_config=input_config, output_config=output_config)
+            query=query, input_config=input_config, output_config=output_config)
         return self.example_gen
 
     @build_step('example_gen')
@@ -219,7 +223,8 @@ class PipelineDef:
 
         Returns: self
         """
-        self.user_schema_importer = input_builders.with_imported_schema(uri)
+        self.user_schema_importer = input_builders.with_imported_schema(
+            uri=uri)
 
         return self.user_schema_importer
 
@@ -305,7 +310,7 @@ class PipelineDef:
 
         Returns: self
         """
-        self.train_base_model = input_builders.with_base_model(uri)
+        self.train_base_model = input_builders.with_base_model(uri=uri)
 
         return self.train_base_model
 
@@ -320,7 +325,7 @@ class PipelineDef:
         Returns: self
         """
         self.user_hyperparameters_importer = input_builders.with_hyperparameters(
-            uri)
+            uri=uri)
 
         return self.user_hyperparameters_importer
 
@@ -344,13 +349,11 @@ class PipelineDef:
         if self.transform:
             args['transform_graph'] = self.transform.outputs['transform_graph']
 
-        if not example_input and self.cached_example_input:
-            args['examples'] = self.cached_example_input
-        elif example_input:
+        if example_input:
             inputs = example_input(self)
             args['examples'] = inputs
 
-            self.cached_example_input = self.cached_example_input or inputs
+            self.cached_example_input = inputs
         else:
             args['examples'] = ExampleInputs.PREPROCESSED_EXAMPLES(
                 self)
@@ -372,6 +375,8 @@ class PipelineDef:
         If there is a user provided schema, it will be used.
         Similarly, if there are hyperparameters explicitly defined, they will be used.
 
+        No need to explicitly re-specify example inputs if you've already specified them in `tune()`.
+
         By default, an trainer_executor.GenericExecutor is used.
 
         If ai platform training arguments are provided, the executor is set to ai_platform_trainer_executor.GenericExecutor
@@ -386,7 +391,7 @@ class PipelineDef:
         """
         args = {
             'module_file': module_file,
-            'schema': SchemaInputs.SCHEMA_CHANNEL(self),
+            'schema': SchemaInputs.SCHEMA_CHANNEL(self)
         }
 
         hparams = HyperParameterInputs.BEST_HYPERPARAMETERS(self)
@@ -440,21 +445,26 @@ class PipelineDef:
                        example_provider_component: Optional[BaseComponent] = None):
         """Constructs an Evaluator component
 
+        No need to specify example_input if you've already specified it in `train()` or `tune()`.
+
         Args:
             example_input (Optional, optional): A `ExampleInputs.{RAW_EXAMPLES, PREPROCESSED_EXAMPLES}` function reference which provides
             the example input channel. Defaults to None, where the input channel is set to the transformed/preprocessed examples. Defaults
-            to RAW_EXAMPLES
+            to RAW_EXAMPLES.
+
 
             example_provider_component (Optional[BaseComponent], optional): An external example input component,
-            which should output examples at the `outputs['examples']` attribute. 
+            which should output examples at the `outputs['examples']` attribute.
 
         Returns: self
         """
 
-        self.lasest_blessed_model_resolver = input_builders.get_latest_blessed_model_resolver()
+        self.latest_blessed_model_resolver = input_builders.get_latest_blessed_model_resolver()
+        self.components['latest_blessed_model_resolver'] = self.latest_blessed_model_resolver
 
         args = {
             'model': self.trainer.outputs['model'],
+            'baseline_model': self.latest_blessed_model_resolver.outputs['model'],
             'eval_config': eval_config
         }
 
@@ -485,6 +495,9 @@ class PipelineDef:
                        example_input=None):
         """Constructs an InfraValidator component.
 
+        No need to specify example_input if you've already specified it in `train()`, `tune()` or `evaluate_model()`.
+
+
         Args:
             example_input ([type], optional): A `ExampleInputs.{RAW_EXAMPLES, PREPROCESSED_EXAMPLES}` function reference which provides
             the example input channel. Defaults to None, where the input channel is set to the transformed/preprocessed examples.
@@ -508,7 +521,8 @@ class PipelineDef:
             inputs = example_input(self)
             args['examples'] = inputs
 
-            self.cached_example_input = self.cached_example_input or inputs
+            # unnecessary to assign `inputs` to `self.cached_example_input` here
+            # this is probably the last stop for the cached example input train
         else:
             args['examples'] = ExampleInputs.RAW_EXAMPLES(
                 self)
@@ -566,7 +580,7 @@ class PipelineDef:
         blessing.
 
         Args:
-            example_provider_component (BaseComponent): A user-provided component that provides example tfrecord inputs to the 
+            example_provider_component (BaseComponent): A user-provided component that provides example tfrecord inputs to the
             `.outputs['examples']` attribute.
 
         Returns: self
@@ -612,11 +626,11 @@ class PipelineDef:
         self.beam_pipeline_args = args
         return self
 
-    def build(self) -> pipeline.Pipeline:
+    def build(self) -> Pipeline:
         """Builds the pipeline.
 
         Returns:
-            pipeline.Pipeline: The native TFX pipeline
+            Pipeline: The native TFX pipeline
         """
         args = {
             'pipeline_name': self.pipeline_name,
@@ -631,23 +645,23 @@ class PipelineDef:
         if self.beam_pipeline_args:
             args['beam_pipeline_args'] = self.beam_pipeline_args
 
-        return pipeline.Pipeline(**args)
+        return Pipeline(**args)
 
 
 class ExampleInputs:
     """Provides accessor functions for channel artifact access, regarding example tf records.
     RAW EXAMPLES -> import example_gen files
-    PREPROCESSED_EXAMPLES -> import transformed examples 
+    PREPROCESSED_EXAMPLES -> import transformed examples
     """
     @staticmethod
-    def _get_raw_examples_channel(pipeline_def: PipelineDef):
-        return pipeline_def.example_gen['examples']
+    def _get_raw_examples_channel(pipeline_def: PipelineDef) -> Channel:
+        return pipeline_def.example_gen.outputs['examples']
 
     @staticmethod
-    def _get_preprocessed_examples_channel(pipeline_def: PipelineDef):
+    def _get_preprocessed_examples_channel(pipeline_def: PipelineDef) -> Channel:
         return pipeline_def.transform.outputs['transformed_examples']
 
-    RAW_EXAMPLES = _get_preprocessed_examples_channel
+    RAW_EXAMPLES = _get_raw_examples_channel
     PREPROCESSED_EXAMPLES = _get_preprocessed_examples_channel
 
 
@@ -657,12 +671,12 @@ class HyperParameterInputs:
     """
     @staticmethod
     def _get_best_hyperparameters(pipeline_def: PipelineDef):
-        if pipeline_def.tuner:
-            return pipeline_def.tuner.outputs['best_hyperparameters']
-
         if pipeline_def.user_hyperparameters_importer:
             return pipeline_def.user_hyperparameters_importer.outputs['result']
 
+        if pipeline_def.tuner:
+            return pipeline_def.tuner.outputs['best_hyperparameters']
+
         return None
 
     BEST_HYPERPARAMETERS = _get_best_hyperparameters
